(* $Id: godi_file.mli 830 2008-06-15 11:37:34Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(* This is an abstraction layer that hides the differences between
   true POSIX and emulated POSIX file systems. On Win32, we emulate the
   Cygwin view on files.

   Note that this is mainly a file name translation module. We cannot
   emulate very special cases that are handled in Cygwin but go beyond
   name translation. For example, there is no translation for
   /dev/zero because a zero device does not exist in Windows.

   Further general restrictions for the Cygwin emulation:
   - No support for ntea or ntsec
   - No support for the special handling of NFS mounts
   - No support for old-style symlinks
   - The Cygwin mount table must exist in the registry

   The [filename] values cache the Win32 translation until the next [chdir].
   Note that there are cases where the cached translation becomes wrong.
   Especially, when a symlink is changed that is part of the original
   file path, this is not automatically reflected. If this can happen in
   the user's code, it is suggested to create filenames only immediately
   before they are used.

   There are probably subtle bugs in the emulation layer, so use it with care.

   All functions indicate errors by throwing the Sys_error exception, except
   when something else is described.
 *)



type filename
  (* On POSIX systems, a filename is just a string.
     On Win32 systems, a filename can either by a Cygwin name or a native
     name. Of course, only the native name can be used to access files,
     so we need some transformation of the Cygwin name to the native name.
     This is done at the time of the first access of the native name, and
     again after [chdir] invocations.
   *)

val filename : string -> filename
  (* Interprets the string as filename. 

     POSIX: The filename can be absolute or relative. There are no restrictions
     on the name.

     Win32: The filename can be:
      1. A relative name. In this case it cannot be recognized by syntactical
         characteristics whether the name is a Cygwin name or a native name.
         (Cygwin names can cross symlinks.) For this reason, relative names
         are treated as Cygwin names, and are subject to the Cygwin resolution.
      2. An absolute name beginning with a slash. This form is taken as
         a Cygwin name.
      3. A drive letter absolute name (c:/dir/file). This form is taken as
         a native name.
      4. A remote name (//host/share/name). This form is taken as a native
         name.
      In all cases, forward and backward slashes are treated as being
      equivalent. "Absolute" names relative to a drive are not supported.
   *)

val inputname : filename -> string
  (* The filename in its original, uninterpreted form *)

val slashname : filename -> string
  (* The filename in its original, uninterpreted form, with the only 
     exception that backslashes are transformed to forward slashes on Win32.
   *)

type nameclass =
    [ `Relative | `POSIX_absolute | `WIN32_drive | `WIN32_remote | `Error ]

val nameclass : filename -> nameclass
  (* Classifies the file name *)

val nativename : ?resolve_final_symlink:bool -> filename -> string
  (* The translated native file name.
     On Win32 this returns a drive-letter or UNC path using slashes as
     component separators.

     resolve_final_symlink: If true (the default), it is tried to resolve the
     final component of the file path as symlink. If false, this is not done.
     Note that you don't necessarily get the symlink itself if you pass
     false here as Cygwin uses the extension ".lnk" for symlinks
   *)

val posixify : filename -> filename
  (* Tries to convert the filename so that Win32-specific name classes are
     avoided
   *)

type split_filename =
  [ `Relative of string list
  | `POSIX_absolute of string list
  | `WIN32_drive of string * string list
  | `WIN32_remote of string * string list
  | `Error
  ]

val split_filename : filename -> split_filename
val combine_filename : split_filename -> filename

val uquote : string -> string
  (* Filename quoting for Unix shells, no matter what the underlying OS
   * is
   *)

val concat : filename -> filename -> filename
  (* If the second name is absolute, the second name is returned. Otherwise
     the names are concatenated, so that the second name is interpreted
     relative to the first name
   *)

val addname : filename -> string -> filename
  (* Add a relative path to the directory (i.e. dir/name). The path uses
     slashes as separators. Special cases:
     - Adding an absolute path like /dir/name does the same as adding
       the corresponding relative path dir/name
     - Adding the empty string leaves the file name unmodified
   *)

val addsuffix : filename -> string -> filename
  (* Add a suffix to the file name *)

val basename : filename -> string
  (* Returns the basename. Returns "/" if the input name is absolute and 
     points to the root of the file system (e.g. for "c:/" just "/" is
     returned).
   *)

val dirname : filename -> filename option
  (* If there is a containing directory name, this name is returned,
     otherwise None. For relative input names, the dirname is "." in doubt.
   *)

val getcwd : unit -> filename option
  (* Returns the physical path of the current directory, if posssible *)

val chdir : filename -> unit
  (* Change to this directory *)

exception Not_a_prefix
val strip_prefix : prefix:split_filename -> split_filename -> string list

type mode = [ `Text | `Binary ]

val open_in : mode -> filename -> in_channel
val open_in_gen : open_flag list -> int -> filename -> in_channel
val open_out : mode -> filename -> out_channel
val open_out_gen : open_flag list -> int -> filename -> out_channel
  (* Open a file as channel. The file is given by a filename *)

val read_dir : ?include_dots:bool -> filename -> string list
  (* Read directory and returns files as list.
   * If [include_dots],  "." and ".." will be returned (default: false)
   *)

val read_dir_hierarchy : ?include_dirs:bool -> filename -> string list
  (* Read the directory like [read_dir], but also descent into subdirectories
   * (but not into symlinks pointing to subdirectories).
   * If [include_dirs], directories are returned (default: false).
   * It is an error if we are not permitted to read subdirectories.
   *
   * For example, if we have the tree
   * d1
   *  +- f1
   *  +- d2
   *  |  +- f2
   *  +- f3
   * read_dir_hierarchy "d1" will return [ "f1"; "d2/f2"; "d3" ], and
   * read_dir_hierarchy ~include_dirs:true "d1" will return
   * [ "f1"; "d2"; "d2/f2"; "d3" ].
   *)

val read_file : mode:mode -> filename -> string
  (* Read the whole file and return it as string. In `Text mode, CR/LF is
   * converted to LF (even on Unix).
   *)

val read_file_lines : filename -> string list
  (* Read the file line by line and return the lines (implies text mode) *)

val write_file : mode:mode -> filename -> string -> unit
  (* Write the whole file. First string is name, second are the contents. 
   * In `Text mode, the file is opened in text mode, whatever this means
   * for the OS.
   *)

val write_file_lines : filename -> string list -> unit
  (* Write the whole file line by line (implies text mode) *)

val file_is_dir : filename -> bool
  (* Whether the file exists and is a directory *)

val file_exists : ?symlink:bool -> filename -> bool
  (* Whether the file exists.
   *
   * @param symlink If true, it is acceptable that the file is a symlink
   * that points to a non-existing location
   *)

val file_type : ?is_regular:bool ->
                ?is_dir:bool ->
		?is_existing:bool ->
                ?is_symlink:bool ->
		?is_not_symlink:bool -> filename -> bool
  (* Checks the file type. The function returns [true] if all of the
   * enabled checks yield a positive result:
   * - [is_regular]: The file exists and is a regular file, or a
   *   symbolic link to an existing regular file
   * - [is_dir]: The file exists and is a directory, or a symbolic link
   *   to an existing directory
   * - [is_existing]: The file exists, or is a symbolic link to
   *   an existing file
   * - [is_symlink]: The file exists and is a symbolic link, no matter
   *   to which location it points to
   * - [is_not_symlink]: The file exists and is not a symbolic link
   *)

val file_mtime : ?symlink:bool -> filename -> float
  (* Returns the time of last modification (or Sys_error).
     symlink: If true, and the file is a symlink, return the mtime of the
     symlink
   *)

val set_mtime : filename -> float -> unit
  (* Sets the mtime (and the atime) of the file. *)

val file_size : filename -> int64
  (* Returns the size of the file *)

val set_mode : filename -> int -> unit
  (* Changes the file mode *)

val stat : ?symlink:bool -> filename -> Unix.LargeFile.stats
  (* Return the stats of the file.
     symlink: If true, and the file is a symlink, return the stats of the
     symlink (lstat)
   *)

val remove : ?recurse:bool -> filename -> unit
  (* Removes the named file. If [recurse], the removal is recursive,
   * and it is no error when the file is not an existing pathname.
   * [Sys_error] when the removal fails.
   *)

val remove_dir : ?recurse:bool -> filename -> unit
  (* Removes the named directory. If [recurse], the removal is recursive,
   * and it is no error when the directory does not exist.
   * [Sys_error] when the removal fails.
   * Unless [recurse] is given, the removal of a non-empty directory fails.
   *)

val rename : filename -> filename -> unit
  (* Renames old to new *)

val create_dir : ?recurse:bool -> filename -> unit
  (* Creates the named directory. If [recurse], it is tried to create the
   * missing parent directories, too, and it is no error when the directory
   * already existed before.
   * [Sys_error] when the creation fails.
   *)

exception Not_a_symlink

val read_symlink : filename -> string
  (* Reads the symlink (Not_a_symlink if not a symlink) *)

val create_symlink : filename -> filename -> unit
  (* [create_symlink src link] creates [link] as symbolic link to [src].

     Win32: This is supported, and Cygwin symlinks are created.
   *)

val create_link : filename -> filename -> unit
  (* [create_link src link] creates [link] as a (hard) link to the same file as
     [src]
   *)

val copy : ?prefer_link:bool -> ?remove:bool -> filename -> filename -> unit
  (* Copies the first file to the second. By default, the second file is
   * overwritten.
   *
   * The permission bits are copied land 0o777. The owner information is
   * not copied, neither are the file times.
   *
   * @param prefer_link When possible, the second file becomes another link
   * of the first file. (Unsupported on Win32)
   *
   * @param remove The second file is not overwritten, but deleted if existing,
   * and again created.
   *)

type file_id

val file_id : ?symlink:bool -> filename -> file_id
  (* Returns an identifier for the file. It is required that the file name
   * f1 and f2 refer to the same file when file_id f1 = file_id f2; however,
   * the reverse need not be true.
   *
   * A trivial implementation is to return a new number each time the
   * function is called. However, this is very bad.
   *
   * For Unix, a file_id is the pair of filesystem device and inode number.
   *
   * symlink: By passing [true] the identity of the symlink itself is returned,
   * not the identity of the linked file.
   *)

val same_file : filename -> filename -> bool
  (* Returns [true] when both files exist and both names refer to the same
   * file. It is ok to return [false] when it cannot be determined whether
   * the files are the same.
   *)

type file_fingerprint

val file_fingerprint : filename -> file_fingerprint
  (* When the fingerprint of the file changes, this is a hint that the
   * contents of the file have changed.
   *
   * For Unix, the fingerprint contains mtime, ctime, and file size.
   *)

val init_tmpdir_state : string -> unit
val get_tmpdir_state : unit -> string
  (* Init/get the tmpdir state *)

val create_tmp_dir : unit -> filename
  (* Creates a new temporary directory. The caller is responsible for deleting
   * it after usage
   *)

type tmpdir
  (* A registered temporary directory *)

val tmpdir_path : tmpdir -> filename
  (* Returns the absolute path to the temporary directory *)

val allocate_tmpdir : unit -> tmpdir
  (* Allocates a new temporary directory *)

val release_tmpdir : tmpdir -> unit
  (* Releases the temporary directory (i.e. deletes it) *)

val release_all_tmpdirs : unit -> unit
  (* Releases all known temporary directories *)


val execute : ?setx:bool -> string -> unit
  (* Execute the shell commands in the string with the "exec" system call.
   * On success, this function does not return. The command string may
   * contain "%s" which is replaced by [get_tmpdir_state()].
   *
   * Note: This is only needed for self-updating packages, and not during
   * the bootstrap.
   *
   * setx: whether to pass the -x option to the shell. Default: false
   *)
